from typing import Literal

from pydantic import RedisDsn, ValidationError, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    # ==================
    # Logging
    # ==================

    #: FastAPI debug mode
    debug: bool = False
    #: Default log level.  Choose from any of the standard Python log levels.
    loglevel: Literal["NOTSET", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"] = "INFO"
    #: What format should we log in?  Valid values are ``json`` and ``text``
    log_type: Literal["json", "text"] = "text"

    # ==================
    # HTTP
    # ==================

    #: Use this as the title for the login form, to give a hint to the
    #: user as to what they're logging into
    auth_realm: str = "Restricted"

    # ==================
    # Session
    # ==================

    #: The name of the cookie to set when a user authenticates
    cookie_name: str = "nginxauth"
    #: The domain to use for our session cookie, if any.
    cookie_domain: str | None = None
    #: The secret key to use for session cookies
    secret_key: str
    #: The maximum age of a session cookie in seconds
    session_max_age: int = 0
    #: Reset the session lifetime to :py:attr:`session_max_age` every time the
    #: user accesses the protected site
    use_rolling_session: bool = False
    #: Session type: either ``redis`` or ``memory``
    session_backend: Literal["redis", "memory"] = "memory"
    #: If using the Redis session backend, the DSN on which to connect to Redis.
    #:
    #: A fully specified Redis DSN looks like this::
    #:
    #:       redis://[username][:password]@host:port/db
    #:
    #: * The username is only necessary if you are using role-based access
    #:   controls on your Redis server.  Otherwise the password is sufficient if you
    #:   have a server password for your Redis server.
    #: * If you don't specify a database, ``0`` is used.
    #: * If you don't specify a password, no password is used.
    #: * If you don't specify a port, ``6379`` is used.
    redis_url: RedisDsn | None = None
    #: If using the Redis session backend, the prefix to use for session keys
    redis_prefix: str = "nginx_ldap_auth."

    # ==================
    # LDAP
    # ==================

    #: The URI via which to connect to LDAP
    ldap_uri: str
    #: The DN as which to bind to LDAP
    ldap_binddn: str
    #: The password to use when binding to LDAP when doing our searches
    ldap_password: str
    #: Whether to use TLS when connecting to LDAP
    ldap_starttls: bool = True
    #: Whether to disable LDAP referrals
    ldap_disable_referrals: bool = False
    #: The base DN under which to perform searches
    ldap_basedn: str
    #: The base DN to append to the user's username when binding.  This is only
    #: important for Active Directory, where we need to use the value of
    #: ``userPrincipalName`` (typically the user's email address) as the
    #: username intead of the dn which would be built as
    #: ``sAMAccountName=user,{LDAP_BASEDN}``.  Include the ``@`` at the begining
    #: of the string.  If this is set, the binddn will be
    #: ``{username}{ldap_user_basedn}``
    ldap_user_basedn: str | None = None
    #: The LDAP attribute to use as the username when searching for a user
    ldap_username_attribute: str = "uid"
    #: The LDAP attribute to use as the full name when getting search results
    ldap_full_name_attribute: str = "cn"
    #: The LDAP search filter to use when searching for a user.  This should
    #: be a valid LDAP search filter.  The search will be a SUBTREE search
    #: with the base DN of :py:attr:`ldap_basedn`.
    #:
    #: You may use these replacement fields in the filter:
    #:
    #: - ``{username_attribute}``: the value of
    #:   :py:class:`Settings.ldap_username_attribute`
    #: - ``{username_full_name_attribute}``: the value of
    #:   :py:class:`Settings.ldap_full_name_attribute`
    #:
    #: Use ``{username}`` in the search filter as the placeholder for the username
    #: supplied by the user from the login form.
    ldap_get_user_filter: str = "{username_attribute}={username}"
    #: The LDAP search filter to use to determine whether a user is authorized.  This
    #: should a valid LDAP search filter. If this is ``None``, all users who can
    #: successfully authenticate will be authorized.  If this is not ``None``,
    #: the search with this filter must return at least one result for the user
    #: to be authorized.
    #:
    #: You may use these replacement fields in the filter:
    #:
    #: - ``{username_attribute}``: the value of
    #:   :py:attr:`ldap_username_attribute`
    #: - ``{username_full_name_attribute}``: the value of
    #:   :py:attr:`ldap_full_name_attribute`
    #:
    #: Use ``{username}`` in the search filter as the placeholder for the username
    #: supplied by the user from the login form.
    ldap_authorization_filter: str | None = None
    #: Number of seconds to wait for an LDAP connection to be established
    ldap_timeout: int = 15
    #: Min number of LDAP connections to keep in the pool
    ldap_min_pool_size: int = 1
    #: Max number of LDAP connections to keep in the pool
    ldap_max_pool_size: int = 30
    #: Recycle LDAP connections after this many seconds
    ldap_pool_connection_lifetime_seconds: int = 20

    # ==================
    # Sentry
    # ==================
    #: The sentry DSN to use for error reporting.  If this is ``None``, no
    #: error reporting will be done.
    sentry_url: str | None = None

    model_config = SettingsConfigDict()

    @model_validator(mode="after")  #: type: ignore
    def redis_url_required_if_session_type_is_redis(self):
        """
        If we've configured the session backend to be ``redis``,
        :py:attr:`redis_url` is required.

        Raises:
            ValidationError: ``redis_url`` is required if ``session_backend`` is
            ``redis``

        """
        if self.session_backend == "redis" and not self.redis_url:
            msg = "redis_url is required if session_backend is redis"
            raise ValidationError(msg)
        return self



import typing

from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import Message, Receive, Scope, Send
from starsessions import SessionMiddleware as StarsessionsSessionMiddleware
from starsessions.middleware import LoadGuard
from starsessions.session import SessionHandler, get_session_remaining_seconds


class SessionMiddleware(StarsessionsSessionMiddleware):
    """
    Override the :py:class:`starsession.SessionMiddleware` to allow us to set
    the cookie name and domain via the ``X-Cookie-Name`` and ``X-Cookie-Domain``
    headers, respectively.  If those headers are not present, the values from
    the constructor are used.

    We need this so that we can set the cookie name and domain dynamically based
    on the request.  This is necessary because we may have multiple nginx severs
    that use a single ``nginx_ldap_auth`` server for authentication.

    Note:
        Unfortunately, the  :py:meth:``__call__`` method is monolithic in the
        superclass, so we have to re-implement it here in is entirety to do
        what we want to do.

    """

    #: The header name for the cookie name passed in by nginx.
    COOKIE_NAME_HEADER: typing.Final[str] = "X-Cookie-Name"
    #: The header name for the cookie domain passed in by nginx.
    COOKIE_DOMAIN_HEADER: typing.Final[str] = "X-Cookie-Domain"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        cookie_name = connection.headers.get(self.COOKIE_NAME_HEADER, self.cookie_name)
        cookie_domain = connection.headers.get(
            self.COOKIE_DOMAIN_HEADER, self.cookie_domain
        )
        session_id = connection.cookies.get(cookie_name)
        handler = SessionHandler(
            connection, session_id, self.store, self.serializer, self.lifetime
        )

        scope["session"] = LoadGuard()
        scope["session_handler"] = handler

        async def send_wrapper(message: Message) -> None:
            if message["type"] != "http.response.start":
                await send(message)
                return

            if not handler.is_loaded:  # session was not loaded, do nothing
                await send(message)
                return

            nonlocal session_id
            path = self.cookie_path or scope.get("root_path", "") or "/"

            if handler.is_empty:
                # if session was initially empty then do nothing
                if handler.initially_empty:
                    await send(message)
                    return

                # session data loaded but empty, no matter whether it was
                # initially empty or cleared we have to remove the cookie and
                # clear the storage
                if not self.cookie_path or (
                    self.cookie_path and scope["path"].startswith(self.cookie_path)
                ):
                    headers = MutableHeaders(scope=message)
                    header_value = "{}={}; {}".format(
                        cookie_name,
                        f"null; path={path}; expires=Thu, 01 Jan 1970 00:00:00 GMT;",
                        self.security_flags,
                    )
                    headers.append("Set-Cookie", header_value)
                    await handler.destroy()
                await send(message)
                return

            # calculate cookie/storage expiry seconds based on selected strategy
            remaining_time = 0

            # if lifetime is zero then don't send max-age at all
            # this will create session-only cookie
            if self.lifetime > 0:
                if self.rolling:
                    # rolling strategy always extends cookie max-age by lifetime
                    remaining_time = self.lifetime
                else:
                    # non-rolling strategy reuses initial expiration date
                    remaining_time = get_session_remaining_seconds(connection)

            # persist session data
            session_id = await handler.save(remaining_time)

            headers = MutableHeaders(scope=message)
            header_parts = [
                f"{cookie_name}={session_id}",
                f"path={path}",
            ]

            username = scope["session"].get("username")
            if username:
                headers.append("X-Authenticated-User", username)

            if self.lifetime > 0:  # always send max-age for non-session scoped cookie
                header_parts.append(f"max-age={remaining_time}")

            if cookie_domain:
                header_parts.append(f"domain={cookie_domain}")

            header_parts.append(self.security_flags)
            header_value = "; ".join(header_parts)
            headers.append("set-cookie", header_value)

            await send(message)

        await self.app(scope, receive, send_wrapper)



from pathlib import Path
from typing import Annotated, Any, cast

from fastapi import Depends, FastAPI, Request, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi_csrf_protect import CsrfProtect
from fastapi_csrf_protect.exceptions import CsrfProtectError
from pydantic import AnyUrl, Field
from pydantic_settings import BaseSettings
from starsessions import (
    InMemoryStore,
    SessionStore,
    get_session_handler,
    load_session,
)
from starsessions.stores.redis import RedisStore

from nginx_ldap_auth import __version__
from nginx_ldap_auth.settings import Settings

from ..logging import get_logger
from .forms import LoginForm
from .middleware import SessionMiddleware
from .models import User

current_dir: Path = Path(__file__).resolve().parent
static_dir: Path = current_dir / "static"
templates_dir: Path = current_dir / "templates"


settings = Settings()

# --------------------------------------
# Session Store
# --------------------------------------

if settings.session_backend == "memory":
    store: SessionStore = InMemoryStore()
    get_logger().info("session.store", backend=settings.session_backend)
elif settings.session_backend == "redis":
    store = RedisStore(
        str(settings.redis_url),
        prefix=settings.redis_prefix,
        gc_ttl=settings.session_max_age,
    )
    redis_url = cast(AnyUrl, settings.redis_url)
    get_logger().info(
        "session.store",
        backend=settings.session_backend,
        server=redis_url.host,
        port=redis_url.port,
        db=redis_url.path,
    )

# --------------------------------------
# CSRF Protection
# --------------------------------------


class CsrfSettings(BaseSettings):
    #: The secret key to use for CSRF tokens
    secret_key: str = Field(validation_alias="CSRF_SECRET_KEY")
    #: We'll set the SameSite attribute on our CSRF cookies to this value
    cookie_samesite: str = "lax"
    #: Set our CSRF cookie to be secure
    cookie_secure: bool = True
    #: Set the maximum age of our CSRF cookie to 5 minutes
    max_age: int = 300
    #: Cookie name
    cookie_key: str = f"{settings.cookie_name}_csrf"
    #: Cookie domain
    cookie_domain: str | None = settings.cookie_domain
    #: Token location for validation -- in the csrf_token field in the body
    token_location: str = "body"  # noqa: S105
    #: The key to use for the CSRF token -- the name of the field in the body
    token_key: str = "csrf_token"  # noqa: S105


@CsrfProtect.load_config
def get_csrf_config():
    return CsrfSettings()


# --------------------------------------
# The FastAPI app
# --------------------------------------

app = FastAPI(title="nginx_ldap_auth", debug=settings.debug, version=__version__)
app.mount("/auth/static", StaticFiles(directory=str(static_dir)), name="static")
app.add_middleware(
    SessionMiddleware,
    store=store,
    cookie_name=settings.cookie_name,
    lifetime=settings.session_max_age,
)
get_logger().info(
    "session.setup.complete",
    backend=settings.session_backend,
    cookie_name=settings.cookie_name,
    cookie_domain=settings.cookie_domain,
    max_age=settings.session_max_age,
    rolling=settings.use_rolling_session,
)
templates = Jinja2Templates(directory=str(templates_dir))


# --------------------------------------
# Startup and Shutdown Events
# --------------------------------------


@app.on_event("startup")
async def startup() -> None:
    """
    Create the LDAP connection pool when we start up.
    """
    await User.objects.create_pool()


@app.on_event("shutdown")
async def shutdown() -> None:
    """
    Close the LDAP connection pool when we shut down.
    """
    await User.objects.cleanup()


# --------------------------------------
# Helper Functions
# --------------------------------------


async def kill_session(request: Request) -> None:
    """
    Kill the current session.

    This means empty the session object of all contents, and delete it from the
    backend.

    Args:
        request: The request object

    """
    for key in list(request.session.keys()):
        del request.session[key]
    await get_session_handler(request).destroy()


# --------------------------------------
# Views
# --------------------------------------


@app.get("/auth/login", response_model=None, name="login")
async def login(
    request: Request,
    csrf_protect: Annotated[CsrfProtect, Depends()],
    service: str = "/",
) -> HTMLResponse | RedirectResponse:
    """
    If the user is already logged in -- they have a session cookie, the session
    the cookie points to exists, and the "username" key in the session exists --
    redirect to the URI named by the ``service``, query paremeter, defaulting to
    ``/`` if that is not present.  Otherwise, render the login page with a fresh
    CSRF token, storing ``service`` in the login form as a hidden field.

    If the header ``X-Auth-Realm`` is set, use that as the title for the login
    page.  Otherwise use
    :py:attr:`nginx_ldap_auth.settings.Settings.auth_realm`.

    Args:
        request: The request object
        csrf_protect: The CSRF protection dependency

    Keyword Args:
        service: redirect the user to this URL after successful login

    Returns:
        If the user is already logged in, a redirect response to the service URL.
        Otherwise, a rendered login page.

    """
    csrf_token, signed_token = csrf_protect.generate_csrf_tokens()
    auth_realm = request.headers.get("x-auth-realm", settings.auth_realm)
    _logger = get_logger(request)
    _logger.info("auth.login.start", target=service)
    await load_session(request)
    if request.session.get("username"):
        session_id = get_session_handler(request).session_id
        _logger.info(
            "auth.login.success.already_logged_in",
            username=request.session["username"],
            session_id=session_id,
            target=service,
        )
        # semgrep-reason:
        #    The service URL is passed in by nginx, and the user cannot directly
        #    reach this URL unless nginx says it needs to, so this is safe.
        # nosemgrep: tainted-redirect-fastapi  # noqa: ERA001
        return RedirectResponse(url=service)
    # semgrep-reason:
    #    The service URL is passed in by nginx, and the user cannot directly
    #    reach this URL unless nginx says it needs to, so this is safe.
    # nosemgrep: tainted-direct-response-fastapi  # noqa: ERA001
    response = templates.TemplateResponse(
        "login.html",
        {
            "request": request,
            "site_title": auth_realm,
            "service": service,
            "csrf_token": csrf_token,
        },
    )
    csrf_protect.set_csrf_cookie(signed_token, response)
    return response


@app.post("/auth/login", response_model=None, name="login_handler")
async def login_handler(
    request: Request,
    csrf_protect: Annotated[CsrfProtect, Depends()],
) -> HTMLResponse | RedirectResponse:
    """
    Process our user's login request.  Validate the CSRF token from the login form,
    and attempt to bind to our LDAP server with the supplied username and password.
    If authentication is successful, redirect to the value of the ``service``
    hidden input field on our form.  If authentication fails, display the login
    form again.

    If the header ``X-Auth-Realm`` is set, use that as the title for the login
    page.  Otherwise use
    :py:attr:`nginx_ldap_auth.settings.Settings.auth_realm`.

    Side Effects:
        If authentication is successful, the user's username is stored in the
        session.

        No matter what, the CSRF cookie is unset to prevent token reuse.

    Args:
        request: The request object
        csrf_protect: The CSRF protection dependency

    Returns:
        A redirect response to the service URL if authentication is successful.
        Otherwise, a rendered login page.

    """
    auth_realm = request.headers.get("x-auth-realm", settings.auth_realm)
    await csrf_protect.validate_csrf(request)
    form = LoginForm(request)
    form.site_title = auth_realm
    await form.load_data()
    if await form.is_valid():
        await load_session(request)
        request.session["username"] = form.username
        response: HTMLResponse | RedirectResponse = RedirectResponse(
            url=form.service, status_code=status.HTTP_302_FOUND
        )
    else:
        response = templates.TemplateResponse("login.html", form.__dict__)
    csrf_protect.unset_csrf_cookie(response)  # prevent token reuse
    return response


@app.get("/auth/logout", response_model=None, name="logout")
async def logout(request: Request) -> RedirectResponse:
    """
    Log the user out by invalidating the sesision, and redirect them to the
    login page.

    Args:
        request: The request object

    Returns:
        A redirect response to the login page.

    """
    _logger = get_logger(request)
    await load_session(request)
    if username := request.session.get("username"):
        await kill_session(request)
        _logger.info("auth.logout", username=username)
    return RedirectResponse(url="/auth/login?service=/")


@app.get("/check")
async def check_auth(request: Request, response: Response) -> dict[str, Any]:
    """
    Ensure the user is still authorized.  If the user is authorized, return
    200 OK, otherwise return 401 Unauthorized.

    The user is authorized if the cookie exists, the session the cookie refers
    to exists, and the ``username`` key in the settings is set.  Additionally,
    the user must still exist in LDAP, and if
    :py:attr:`nginx_ldap_auth.settings.Settings.ldap_authorization_filter` is
    not ``None``, the user must also match the filter.

    Side Effects:
        If the user is not authorized, the session is destroyed, and the user is
        status_code on ``response`` is set to 401.

    Args:
        request: The request object
        response: The response object

    Returns:
        An empty dictionary.

    """
    if request.cookies.get(settings.cookie_name):
        await load_session(request)
        if request.session.get("username"):
            # We have a valid session
            if not await User.objects.get(request.session["username"]):
                # The user does not exist in LDAP; log them out
                await kill_session(request)
            if not await User.objects.is_authorized(request.session["username"]):
                # The user is no longer authorized; log them out
                await kill_session(request)
            return {}
        # Destroy the session because it is not valid
        await kill_session(request)
    # Force the user to authenticate
    response.headers["Cache-Control"] = "no-cache"
    response.status_code = status.HTTP_401_UNAUTHORIZED
    return {}


@app.exception_handler(CsrfProtectError)
def csrf_protect_exception_handler(request: Request, exc: CsrfProtectError) -> Response:
    """
    Handle CSRF protection errors.  All we're going to do is redirect the user
    back to the login page after logging the error.

    Args:
        request: The request object
        exc: The exception object from the CSRF protection middleware

    Returns:
        A redirect response to the login page.

    """
    _logger = get_logger(request)
    _logger.error("auth.login.csrf.error", error=str(exc))
    return RedirectResponse(
        url=app.url_path_for("login"), status_code=status.HTTP_302_FOUND
    )



import logging  # noqa: A005
import logging.config
import sys
from typing import Any, cast

import structlog
from fastapi import Request
from starlette.datastructures import Address

from nginx_ldap_auth.settings import Settings

settings = Settings()
logger = structlog.get_logger("nginx_ldap_auth")


def get_logger(request: Request | None = None) -> structlog.BoundLogger:
    """
    Return a structlog logger.  If a request is passed, bind the following
    values to the logger:

    * The name of the authentication realm (the title of the login window)
    * The host name of the server that asked us to authenticate/authorize
    * The IP address of the client that wants to authenticate/authorize

    Args:
        request: The request object

    Returns:
        A structlog logger

    """
    if request is None:
        return logger
    remote_ip = request.headers.get("x-forwarded-for")
    if remote_ip:
        remote_ip = remote_ip.split(",")[0]
    else:
        remote_ip = cast(Address, request.client).host
    request = cast(Request, request)
    return logger.bind(
        realm=request.headers.get("x-auth-realm", settings.auth_realm),
        host=request.headers.get("host", "unknown"),
        remote_ip=remote_ip,
    )


class ContextLoggingProcessor:
    def __init__(self, **kwargs: Any) -> None:
        self.log_kwargs = kwargs

    def __call__(self, _, __, event_dict):
        """
        Adds extra runtime event info to our log messages based on what was
        passed to our constructor.

        Does not overwrite any event info that's already been set in the logging
        call.
        """
        for k, v in self.log_kwargs.items():
            event_dict.setdefault(k, v)
        return event_dict


class CensorPasswordProcessor:
    def __init__(self, **kwargs: Any) -> None:
        self.log_kwargs = kwargs

    def __call__(self, _, __, event_dict):
        """
        Automatically censors any logging context key called "password",
        "password1", or "password2".
        """
        for password_key_name in ("password", "password1", "password2"):
            if password_key_name in event_dict:
                event_dict[password_key_name] = "*CENSORED*"
        return event_dict


structlog.configure(
    processors=[
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        ContextLoggingProcessor(),
        CensorPasswordProcessor(),
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=False,
)

pre_chain = [
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.processors.TimeStamper(fmt="iso"),
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "handlers": {
        "default": {
            "level": settings.loglevel,
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "access": {"class": "logging.StreamHandler", "formatter": "access"},
    },
    "loggers": {
        "uvicorn.error": {
            "level": settings.loglevel,
            "handlers": ["default"],
            "propagate": False,
        },
        "uvicorn.access": {
            "level": settings.loglevel,
            "handlers": ["access"],
            "propagate": False,
        },
    },
    "root": {
        # Set up the root logger.  This will make all otherwise unconfigured
        # loggers log through structlog processor.
        "handlers": ["default"],
        "level": settings.loglevel,
    },
    "formatters": {
        "default": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.processors.JSONRenderer(),
            "foreign_pre_chain": pre_chain,
            "format": "%(message)s",
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "format": "%(asctime)s %(message)s",
        },
    },
}


if settings.log_type == "text":
    LOGGING["formatters"]["default"]["processor"] = structlog.dev.ConsoleRenderer()  # type: ignore[index]


logging.config.dictConfig(LOGGING)


def handle_exception(exc_type, exc_value, exc_traceback):
    """
    Log any uncaught exception instead of letting it be printed by Python
    (but leave KeyboardInterrupt untouched to allow users to Ctrl+C to stop)
    See https://stackoverflow.com/a/16993115/3641865
    """
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception
